/*NOTES
This program reads in a stereo hrtf file recorded in (left/right) sample formats as 
16 bit ints.  It calculates an inverse filter based on the right channel hrtf, then
applies the inverse filter to the left channel.  The sampling rate is reduced from 
the original 44100 to 11025 by decimation after low pass filtering in the frequency
domain.

FIXME  This program still needs to be converted from using 
four1.c to using the new C_FFT.c - four1.c is not compatible
with the GPL

NOTE since this was written as a one time utility, it has poor 
programming style.

Karen Bryden 
Manager, Voice Systems and Technologies 


E-mail: karen.bryden@crc.ca



Copyright (C) 2000 Her Majesty the Queen in Right of Canada
                   Communications Research Centre Canada
                   An agency of Industry Canada.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

See File COPYING.txt




*/
#include <stdio.h>
#include <string.h>
//#include <Files.h>
#include <math.h>
#include <stdlib.h>

#define maxTypedChars 80
#define transformSize 8192
#define hrtfSize 128
#define passBand 5400.0
#define stopBand 5512.5 /* need for 4:1 decimation */
#define samplingFrequency 44100.0

int getData ( FILE *, int , int * );
int writeData ( FILE *, int , int * );
void userInput (char *);
void four1(float *, int, int);

void main(void)

{
FILE *impulseResponseFile, *inverseFilterFile;

char searchString[maxTypedChars];

int i, j, result, *int_IR, *temp, 
	Fmax = (int)(transformSize * passBand / samplingFrequency);

float *float_RC, *float_LC, magnitude, phase, attenuation,
	Fstop = (transformSize * stopBand / samplingFrequency);

/* allocate heap space */

/* "transformSize" number of ints */

int_IR = (int *)malloc(sizeof(int) * transformSize);
if (int_IR == NULL)
	{
	printf("Can't allocate space for integer array!");
	return;
	}

/* "transformSize" number of temp ints */

temp = (int *)malloc(sizeof(int) * transformSize);
if (temp == NULL)
	{
	printf("Can't allocate space for integer array!");
	free (int_IR);
	return;
	}

/* TWO times "transformSize" number of floats for right channel */

float_RC = (float *)malloc(sizeof(float) * 2 * transformSize);
if (float_RC == NULL)
	{
	printf("Can't allocate space for R float array!");
	free (int_IR);
	free (temp);
	return;
	}

/* TWO times "transformSize" number of floats for left channel */

float_LC = (float *)malloc(sizeof(float) * 2 * transformSize);
if (float_LC == NULL)
	{
	printf("Can't allocate space for L float array!");
	free (float_RC);
	free (int_IR);
	free (temp);
	return;
	}

/* open files */

printf ("\nenter source filename -->");
userInput(searchString);
impulseResponseFile = fopen( searchString, "rb" );
if ( impulseResponseFile == NULL )
	{
	printf ( "can't find impulse response file!");
	free (float_RC);
	free (float_LC);
	free (int_IR);
	free (temp);
	return;
	}

printf ("\nenter result filename -->");
userInput(searchString);
inverseFilterFile = fopen( searchString, "wb" );
if ( inverseFilterFile == NULL )
	{
	printf ( "can't open inverse filter file!");
	free (float_RC);
	free (float_LC);
	free (int_IR);
	free (temp);
	fclose (impulseResponseFile);
	return;
	}

/* read in the stereo impulse response for left and right channel */

result = getData ( impulseResponseFile, 2 * hrtfSize, int_IR);
if (result == 0)
	{
	printf("Error reading impulse response from file!");
	free (float_RC);
	free (float_LC);
	free (int_IR);
	fclose (impulseResponseFile);
	fclose (inverseFilterFile);
	free (temp);
	return;
	}

/*
Fill the real elements of the fft input array with the data
for the right channel.  Set the imaginary elements to 0.
*/
for (i = 0 ; i < transformSize; i++)
	{
	if (i < hrtfSize) float_RC[2 * i] = (float)int_IR[2 * i + 1] / 16384.0f;
	else float_RC[2 * i] = 0.0;
	float_RC[2 * i + 1] = 0.0;
	}

/*
Fill the real elements of the fft input array with the data
for the left channel.  Set the imaginary elements to 0.
*/
for (i = 0 ; i < transformSize; i++)
	{
	if (i < hrtfSize) float_LC[2 * i] = (float)int_IR[2 * i] / 16384.0f;
	else float_LC[2 * i] = 0.0;
	float_LC[2 * i + 1] = 0.0;
	}

/* call to four1
Data values input to four1 are complex numbers represented
as real followed by imaginary.  Second argument nn is half
the number of data points since nn represents the number of
*complex* data values.  Since four1 uses subscripts running
from 1 to N (not 0 to N-1), must unput the data pointer -1.
Third argument is 1 for fft, -1 for inverse fft.  Inverse fft
has no 1/N scaling, thus its output must be divided by nn.
*/

four1(float_RC - 1, transformSize, 1);
four1(float_LC - 1, transformSize, 1);

/* convert spectrum of right channel to polar coordinates */
for (i = 0 ; i < transformSize; i++)
	{
	magnitude = (float)sqrt(float_RC[2 * i] * float_RC[2 * i] +
						float_RC[2 * i + 1] * float_RC[2 * i + 1]);
	phase = (float)atan2(float_RC[2 * i + 1], float_RC[2 * i]);
	float_RC[2 * i] = magnitude;
	float_RC[2 * i + 1] = phase;
	}

/* convert spectrum of left channel to polar coordinates */
for (i = 0 ; i < transformSize; i++)
	{
	magnitude = (float)sqrt(float_LC[2 * i] * float_LC[2 * i] +
						float_LC[2 * i + 1] * float_LC[2 * i + 1]);
	phase = (float)atan2(float_LC[2 * i + 1], float_LC[2 * i]);
	float_LC[2 * i] = magnitude;
	float_LC[2 * i + 1] = phase;
	}

/* invert the frequency response of the left channel, Ml = Ml / Mr, Pl  = Pl - Ph */

for (i = 0 ; i < transformSize; i++) /* retain DC component */
	{
	if (i <= transformSize / 2) j = i;
	else j = transformSize - i;
	attenuation = (Fstop - j) / (Fstop - Fmax);
	if (attenuation > 1.0) attenuation = 1.0;
	if (attenuation < 0.0) attenuation = 0.0;
	if(float_RC[2 * i] < 0.01)float_RC[2 * i] = 0.01f;
/*	inverse filter the left channel with the right channel frequency response
This transformation applied to the right channel would yield a delta fn */
	float_LC[2 * i] = float_LC[2 * i] / float_RC[2 * i] * attenuation;
	float_LC[2 * i + 1] = float_LC[2 * i + 1] - float_RC[2 * i + 1];
	}

/* convert back to linear coordinates */

for (i = 0 ; i < transformSize; i++)
	{
	magnitude = float_LC[2 * i];
	phase = float_LC[2 * i + 1];
	float_LC[2 * i] = (float)(magnitude * cos(phase));
	float_LC[2 * i + 1] = (float)(magnitude * sin(phase));
	}

/* inverse transform the left channel */

four1(float_LC - 1, transformSize, -1);

/* copy the inverse filter back to the fixed point array */

for (i = 0 ; i < transformSize; i++)
	{
	float_LC[2 * i] = float_LC[2 * i] / transformSize;
	int_IR[i] = (int)(float_LC[2 * i] * 16384.0f);
	}

/* rotate the vector a bit */

for (i = 0 ; i < transformSize; i++)
	{
	j = i + 8080;
	if (j >= transformSize) j -= transformSize;
	temp[i] = int_IR[j];
	}

/* write the first 64 points to the specified file with 4:1 decimation */
for (i = 112; i < 172; i+=4)
	{
	fprintf (inverseFilterFile, "%7.4f, ", (float)temp[i] / 4054.0);
	}

/*writeData (inverseFilterFile, transformSize, temp);*/

free (float_RC);
free (float_LC);
free (int_IR);
free (temp);
fclose (impulseResponseFile);
fclose (inverseFilterFile);

printf("\ntype <cr> to close window \a");
}
