#include <math.h>
#include "defs.h"

/*
C_FFT.c 


Karen Bryden 
Manager, Voice Systems and Technologies 


E-mail: karen.bryden@crc.ca



Copyright (C) 2000 Her Majesty the Queen in Right of Canada
                   Communications Research Centre Canada
                   An agency of Industry Canada.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

See File COPYING.txt

This function is based on a paper entitled 
"A New Principle for Fast Fourier Transformation", 
Charles M. Rader and N.M. Brenner, I.E.E.E. 
Transactions on Acoustics, Speech and Signal 
Processing, June 1976, pp 264-266.  It is 
loosely based on a translation of a FORTRAN 
subroutine "FOUREA", which appeared in the 
book "Programs for Digital Signal Processing", 
published by I.E.E.E. Press, chapter 1, 
section 1.1, 1979.

*/

void C_FFT(struct cpxFloat *data, int N, int ISI)
{
int i, j, m, mmax, istep;
struct cpxFloat cfTemp;
struct cpxDouble cdTemp1, cdTemp2;
double theta, pi = 3.14159265358979323, dTemp;

//first operation puts data in bit-reversed order
j = 0;
for(i = 0; i < N; i++)
	{
if(i < j)
		{
cfTemp.re = data[j].re;
cfTemp.im = data[j].im;

data[j].re = data[i].re;
data[j].im = data[i].im;

data[i].re = cfTemp.re;
data[i].im = cfTemp.im;
		}
m = N / 2;
while(j >= m)
		{
j = j - m;
m = m / 2;
if(m == 0)break;
		}
j = j + m;
	}

//second operation computes the butterflies
mmax = 1;
while (mmax < N)
	{
istep = 2 * mmax;
theta = pi * ISI / mmax;
dTemp = sin(theta / 2.0);
cdTemp2.re = -2.0 * dTemp * dTemp;
cdTemp2.im = sin(theta);
cdTemp1.re = 1.0;
cdTemp1.im = 0.0;
for(m = 0; m < mmax; m++)
		{
for(i = m; i < N; i += istep)
			{
j = i + mmax;
cfTemp.re = (float)(cdTemp1.re * data[j].re - cdTemp1.im * data[j].im);
cfTemp.im = (float)(cdTemp1.re * data[j].im + cdTemp1.im * data[j].re);
data[j].re = data[i].re - cfTemp.re;
data[j].im = data[i].im - cfTemp.im;
data[i].re += cfTemp.re;
data[i].im += cfTemp.im;
			}
dTemp = cdTemp1.re;
cdTemp1.re = cdTemp1.re * cdTemp2.re - cdTemp1.im * cdTemp2.im + cdTemp1.re;
cdTemp1.im = cdTemp1.im * cdTemp2.re + dTemp * cdTemp2.im + cdTemp1.im;
		}
mmax = istep;
	}
}
